#include <algorithm>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <vector>

#include <rlcsa/rlcsa.h>

#include "doclist.h"
#include "pdltopk.h"


void printUsage();
void bruteQuery(const RLCSA& rlcsa, std::vector<pair_type>& ranges, const std::string& base_name, usint k, usint repet, bool docs);
void pdlQuery(const RLCSA& rlcsa, std::vector<pair_type>& ranges, const std::string& base_name, usint k, usint repet, bool freqs);
void mergeQuery(const RLCSA& rlcsa, std::vector<pair_type>& ranges, const std::string& base_name, usint k, usint repet, PDLTopk::merge_type strategy);

int
main(int argc, char** argv)
{
  std::cout << "Topk-k document retrieval test" << std::endl;
  if(argc < 4)
  {
    printUsage();
    return 1;
  }

  // Parameters
  char* base_arg = 0;
  char* pattern_arg = 0;
  usint k = 0, repet = 1;
  bool brute = false, brute_d = false, pdl = false, pdlfreq = false, merge_brute = false, merge_exp = false, merge_lin = false, selected = false;
  for(int arg = 1; arg < argc; arg++)
  {
    if(argv[arg][0] == '-')
    {
      switch(argv[arg][1])
      {
        case 'b':
          brute = true; selected = true; break;
        case 'B':
          brute_d = true; selected = true; break;
        case 'e':
          merge_exp = true; selected = true; break;
        case 'f':
          pdlfreq = true; selected = true; break;
        case 'l':
          merge_lin = true; selected = true; break;
        case 'm':
          merge_brute = true; selected = true; break;
        case 'p':
          pdl = true; selected = true; break;
        default:
          std::cerr << "Invalid parameter: " << argv[arg] << std::endl;
          break;
      }
    }
    else if(base_arg == 0)    { base_arg = argv[arg]; }
    else if(pattern_arg == 0) { pattern_arg = argv[arg]; }
    else if(k == 0)           { k = atoi(argv[arg]); }
    else { repet = atoi(argv[arg]); }
  }
  if(k == 0)
  {
    printUsage();
    return 1;
  }
  if(!selected) { brute = true; }

  // Output
  std::cout << "Algorithms:";
  if(brute) { std::cout << " brute-l"; }
  if(brute_d) { std::cout << " brute-d"; }
  if(pdl)  { std::cout << " pdl"; }
  if(pdlfreq) { std::cout << " pdl+freqs"; }
  if(merge_brute) { std::cout << " merge_brute"; }
  if(merge_lin) { std::cout << " merge_lin"; }
  if(merge_exp) { std::cout << " merge_exp"; }
  std::cout << std::endl;
  std::string base_name = base_arg;
  std::cout << "Base name: " << base_name << std::endl;
  std::string patterns_name = pattern_arg;
  std::cout << "Patterns: " << patterns_name << std::endl;
  std::cout << "k: " << k << std::endl;
  std::cout << "Repetitions: " << repet << std::endl;
  std::cout << std::endl;

  // Index
  RLCSA rlcsa(base_name);
  if(!(rlcsa.isOk())) { return 2; }
  rlcsa.printInfo();
  rlcsa.reportSize(true);

  // Patterns
  std::ifstream pattern_file(patterns_name.c_str(), std::ios_base::binary);
  if(!pattern_file)
  {
    std::cerr << "Error opening pattern file!" << std::endl;
    return 3;
  }
  std::vector<std::string> rows;
  CSA::readRows(pattern_file, rows, true);
  pattern_file.close();

  // Counting
  std::vector<pair_type> ranges; ranges.reserve(rows.size());
  usint total_size = 0, total_occ = 0;
  double start = CSA::readTimer();
  for(usint i = 0; i < rows.size(); i++)
  {
    pair_type result = rlcsa.count(rows[i]);
    total_size += rows[i].length(); total_occ += CSA::length(result);
    ranges.push_back(result);
  }
  double seconds = CSA::readTimer() - start;
  double megabytes = total_size / (double)CSA::MEGABYTE;
  std::cout << "Patterns:     " << rows.size() << " (" << megabytes << " MB)" << std::endl;
  std::cout << "Occurrences:  " << total_occ << std::endl;
  std::cout << "Counting:     " << seconds << " seconds (" << (megabytes / seconds) << " MB/s, "
                                << (rows.size() / seconds) << " patterns/s)" << std::endl;
  std::cout << std::endl;

  // Brute force
  if(brute) { bruteQuery(rlcsa, ranges, base_name, k, repet, false); }

  // Brute force with DA
  if(brute_d) { bruteQuery(rlcsa, ranges, base_name, k, repet, true); }

  // PDL without frequencies
  if(pdl) { pdlQuery(rlcsa, ranges, base_name, k, repet, false); }

  // PDL with frequencies
  if(pdlfreq) { pdlQuery(rlcsa, ranges, base_name, k, repet, true); }

  // PDL using brute-force merging strategy
  if(merge_brute) { mergeQuery(rlcsa, ranges, base_name, k, repet, PDLTopk::merge_brute); }

  // PDL using linear merging strategy
  if(merge_lin) { mergeQuery(rlcsa, ranges, base_name, k, repet, PDLTopk::merge_lin); }

  // PDL using exponential merging strategy
  if(merge_exp) { mergeQuery(rlcsa, ranges, base_name, k, repet, PDLTopk::merge_exp); }

  return 0;
}

void
printUsage()
{
  std::cout << "Usage: list_topk [algorithms] base_name patterns k [repetitions]" << std::endl;
  std::cout << "  -b  Brute force using locate() (default)" << std::endl;
  std::cout << "  -B  Brute force using D" << std::endl;
  std::cout << "  -e  PDL with exponential merging" << std::endl;
  std::cout << "  -f  PDL with frequencies" << std::endl;
  std::cout << "  -l  PDL with linear merging" << std::endl;
  std::cout << "  -m  PDL with brute-force merging" << std::endl;
  std::cout << "  -p  PDL without frequencies" << std::endl;
  std::cout << std::endl;
}

void
bruteQuery(const RLCSA& rlcsa, std::vector<pair_type>& ranges, const std::string& base_name, usint k, usint repet, bool docs)
{
  CSA::ReadBuffer* docarray = (docs ? readDocarray(rlcsa, base_name) : 0);
  if(docs && docarray == 0) { return; }

  double start = CSA::readTimer();
  usint docc = 0;
  for(usint j = 0; j < repet; j++)
  {
    for(usint i = 0; i < ranges.size(); i++)
    {
      std::vector<Document>* result = (docs ? bruteForceTopk(*docarray, ranges[i], k) : bruteForceTopk(rlcsa, ranges[i], k));
      if(result != 0)
      {
        docc += result->size();
        delete result; result = 0;
      }
    }
  }
  double seconds = CSA::readTimer() - start;

  std::string algorithm = (docs ? "Brute-D" : "Brute-L");
  usint doclist_size = (docs ? docarray->reportSize() : 0);
  printResults(algorithm, seconds, docc, repet * ranges.size());
  printSize(algorithm, rlcsa.getSize(), rlcsa.reportSize(), doclist_size);
  std::cout << std::endl;

  delete docarray; docarray = 0;
}

void
pdlQuery(const RLCSA& rlcsa, std::vector<pair_type>& ranges, const std::string& base_name, usint k, usint repet, bool freqs)
{
  PDLTopk pdltopk(rlcsa, base_name, false, false);
  if(!(pdltopk.isOk())) { return; }

  double start = CSA::readTimer();
  usint docc = 0;
  for(usint j = 0; j < repet; j++)
  {
    for(usint i = 0; i < ranges.size(); i++)
    {
      std::vector<Document>* result = pdltopk.query(ranges[i], k, freqs);
      if(result != 0)
      {
        docc += result->size();
        delete result; result = 0;
      }
    }
  }
  double seconds = CSA::readTimer() - start;

  std::string algorithm = (freqs ? "PDL-freq" : "PDL-nofreq");
  printResults(algorithm, seconds, docc, repet * ranges.size());
  printSize(algorithm, rlcsa.getSize(), rlcsa.reportSize(), pdltopk.reportSize(freqs));
  std::cout << std::endl;
}

void
mergeQuery(const RLCSA& rlcsa, std::vector<pair_type>& ranges, const std::string& base_name, usint k, usint repet, PDLTopk::merge_type strategy)
{
  PDLTopk pdltopk(rlcsa, base_name, false, true);
  if(!(pdltopk.isOk())) { return; }

  double start = CSA::readTimer();
  usint docc = 0;
  for(usint j = 0; j < repet; j++)
  {
    for(usint i = 0; i < ranges.size(); i++)
    {
      std::vector<Document>* result = pdltopk.query(ranges[i], k, strategy);
      if(result != 0)
      {
        docc += result->size();
        delete result; result = 0;
      }
    }
  }
  double seconds = CSA::readTimer() - start;

  std::string algorithm = "PDL-";
  if(strategy == PDLTopk::merge_brute) { algorithm += "brute"; }
  else if(strategy == PDLTopk::merge_lin) { algorithm += "lin"; }
  else if(strategy == PDLTopk::merge_exp) { algorithm += "exp"; }
  printResults(algorithm, seconds, docc, repet * ranges.size());
  printSize(algorithm, rlcsa.getSize(), rlcsa.reportSize(), pdltopk.reportSize(true));
  std::cout << std::endl;
}
