#ifndef REPAIR_H
#define REPAIR_H

#include <cstdio>
#include <fstream>
#include <map>
#include <vector>

#include <rlcsa/misc/utils.h>


using CSA::usint;
using CSA::pair_type;


class RePairEncoder
{
  public:
    typedef uint                            item_type;
    typedef std::vector<item_type>          set_type;
    typedef std::vector<set_type>           set_of_sets;
    typedef std::pair<item_type, item_type> rule_type;
    typedef std::map<rule_type, usint>      map_type;

    const static usint DEFAULT_MIN_FREQ = 10;
    const static usint DEFAULT_MAX_PAIRS = 0x10000000;
    const static usint DEFAULT_MAX_SET = 1024;  // Larger sets are not used to find pairs.

    // The sets are assumed to be sorted in increasing order. They are destroyed during the construction.
    // max_pairs == 0 denotes no limit to the number of pairs considered.
    RePairEncoder(set_of_sets& input, bool print = false, usint min_freq = DEFAULT_MIN_FREQ, usint max_pairs = DEFAULT_MAX_PAIRS, usint max_set = DEFAULT_MAX_SET);
    ~RePairEncoder();

    // This uses similar conventions as Gonzalo's implementation of RePair.
    void writeTo(const std::string& base_name) const;

    inline bool isOk() const { return this->ok; }

    // Compress a new set using the existing grammar.
    // The items of the new set should be less than alphabet_size.
    bool addSet(set_type& new_set);

  private:
    struct GrammarRule
    {
      item_type id;
      usint     freq;
      rule_type items;

      GrammarRule(item_type _id, usint _freq, rule_type _items) : id(_id), freq(_freq), items(_items) {}
    };

    struct GrammarRuleComparator
    {
      bool operator() (const GrammarRule& a, const GrammarRule& b)
      {
        return (a.freq > b.freq); // Decreasing order.
      }
    };

    struct RuleLayer
    {
      std::vector<GrammarRule> rules;
      std::vector<item_type> item_to_rule;
    };

    typedef std::vector<GrammarRule> rule_layer;
    typedef std::vector<RuleLayer*>  grammar_rules;

    const static usint PARALLEL_SORT_MINIMUM = 65536;

    bool ok;

    usint sets, items;
    usint alphabet_size, number_of_rules, rule_layers;

    grammar_rules rules;
    set_of_sets   stored_sets;

    void getStatistics(const set_of_sets& sets);

    usint newAlphabetSize() const { return this->alphabet_size + this->number_of_rules; }
    bool pairsCanExceed(usint limit) const;

    // Returns the number of applied rules.
    usint applyRules(usint layer, set_type& target) const;

    // These are not allowed.
    RePairEncoder();
    RePairEncoder(const RePairEncoder&);
    RePairEncoder& operator = (const RePairEncoder&);
};


#endif  // REPAIR_H
